// MorphingTest.java
package morphingtool;

import java.awt.Point;
import java.io.File;
import java.io.IOException;
import javax.swing.*;

public class MorphingTest {

  // Drei einzelne Fenster
  private JFrame frameImage1;
  private JFrame frameImage2;
  private JFrame frameMorphing;
  // Zwei ImageContainer und ein MorphingContainer für die drei Fenster
  private ImageContainer imageContainer1;
  private ImageContainer imageContainer2;
  private MorphingContainer morphingContainer;
  // Zwei Bilder zum testen
  private File imageContainer1File;
  private File imageContainer2File;
  // Schieberegler Konstante
  private final int SCHIEBEREGLER_WERT_1 = 100;
  private final int SCHIEBEREGLER_WERT_2 = 0;

  public MorphingTest() {
    // Drei einzelne Fenster mit Namen
    frameImage1 = new JFrame("ImageContainer1");
    frameImage2 = new JFrame("ImageContainer2");
    frameMorphing = new JFrame("MorphingContainer");
    // Um die Punktanzeige zu testen, werden alle Punkt Booleans auf den Wert true
    // gesetzt. Die übergebenen Punkte entsprechen den Augen- und Mundpunkten
    // der Bilder "SteveJobs.jpg" und "BillGates.jpg".
    imageContainer1 = new ImageContainer(true, true, true, new Point(144, 142),
            new Point(219, 140), new Point(186, 222));
    imageContainer2 = new ImageContainer(true, true, true, new Point(117, 143),
      new Point(191, 140), new Point(158, 221));
    /* Wichtig: Beim Zusammenbau der Software in der Softwarrunde verwenden
     * Sie nicht die beiden obigen Linien, sondern die beiden untenstehenden
     * auskommentierten. Die Punkte werden vom GUI via Setter Methoden übertragen. */
    // imageContainer1 = new ImageContainer();
    // imageContainer2 = new ImageContainer();
    morphingContainer = new MorphingContainer();
    // Zwei Bilder zum testen.
    // Im Grafik Netbeans Projektordner finden Sie weitere Bilder.
    imageContainer1File = new File("SteveJobs.jpg");
    imageContainer2File = new File("BillGates.jpg");
    // Öffne zwei ImageContainer und ein MorphingContainer je in ein Fenster
    dreiFensterFuerDieContainerOeffnen();
    // Lade je ein Bild in die ImageContainer
    ladeBilderInImageBuffer();
    // Rufe Morphing mit Testwert auf
    morphingContainer.erzeugeMorphingBild(imageContainer1.getBufferedImage(),
      imageContainer2.getBufferedImage(), imageContainer1.getPositionLinkesAuge(),
      imageContainer1.getPositionRechtesAuge(), imageContainer1.getPositionMund(),
      imageContainer2.getPositionLinkesAuge(), imageContainer2.getPositionRechtesAuge(),
      imageContainer2.getPositionMund(), SCHIEBEREGLER_WERT_1, SCHIEBEREGLER_WERT_2);
  }

  private void dreiFensterFuerDieContainerOeffnen() {
    // Fenster schön darstellen
    JFrame.setDefaultLookAndFeelDecorated(true);
    // Knopf zum Schliessen "aktivieren"
    frameImage1.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    frameImage2.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    frameMorphing.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    // Container den Fenstern zuordnen.
    frameImage1.getContentPane().add(imageContainer1);
    frameImage2.getContentPane().add(imageContainer2);
    frameMorphing.getContentPane().add(morphingContainer);
    // Grösse der Fenster an die Container anpassen
    frameImage1.pack();
    frameImage2.pack();
    frameMorphing.pack();
    // Fenster nebeneinander anzeigen lassen
    frameImage1.setLocation(0, 0);
    frameMorphing.setLocation(320, 0);
    frameImage2.setLocation(640, 0);
    // Fenster anzeigen lassen
    frameImage1.setVisible(true);
    frameImage2.setVisible(true);
    frameMorphing.setVisible(true);
  }

  private void ladeBilderInImageBuffer() {
    // Da beim Öffnen von Files unvorhersehbare Fehler entstehen können,
    // müssen eventuelle Fehler mit einem try-catch Block abgefangen werden
    try {
      // Der Hinweis an den Bentuzer wird in diesem Expertenteil nicht in das GUI
      // sondern in das Output Fenster des NetBeans geschrieben
      System.out.println(imageContainer1.ladeBildInImageBuffer(imageContainer1File));
      System.out.println(imageContainer2.ladeBildInImageBuffer(imageContainer2File));
    } catch (IOException ex) {
      // Schreibe eventuelle Fehler in NetBeans Output Fenster, beende Programm
      System.out.println(ex);
      System.exit(1);
    }
  }

  public static void main(String[] args) {
    new MorphingTest();
  }
}

