// MorphingContainer.java
package morphingtool;

import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;

public class MorphingContainer extends ImageContainer {

  private BufferedImage imageBuffer1,  imageBuffer2;
  private AffineTransform transformationBild1,  transformationBild2;
  // Die Koordinaten der Augen- und Mundpunkte werden in den gesichtsDreieck
  // Arrays abgelegt. Da der Ausdruck "gesichtsDreieck1[4]" nicht wirklich
  // aussagekräftig ist, wird "gesichtsDreieck1[MUND_X]" geschrieben, damit klar
  // wird, dass damit die x-Koordinate des Mund Punktes des Bildes 1 gemeint
  // ist. Die Konstante MUND_X trägt daher den Wert 4. Der Typ double[] wurde
  // gewählt, da die affinen Transformationen double Werte verlangen.
  private final int GESICHTS_DREIECK_GROESSE = 6;
  private double[] gesichtsDreieck1 = new double[GESICHTS_DREIECK_GROESSE];
  private double[] gesichtsDreieck2 = new double[GESICHTS_DREIECK_GROESSE];
  private final int LINKES_AUGE_X = 0;
  private final int LINKES_AUGE_Y = 1;
  private final int RECHTES_AUGE_X = 2;
  private final int RECHTES_AUGE_Y = 3;
  private final int MUND_X = 4;
  private final int MUND_Y = 5;

  // Constructor
  public MorphingContainer() {
    //  Übernehme Startbild der Vaterklasse
    super();
  }

  public void erzeugeMorphingBild(BufferedImage bild1, BufferedImage bild2,
          Point positionLinkesAuge1, Point positionRechtesAuge1,
          Point positionMund1, Point positionLinkesAuge2,
          Point positionRechtesAuge2, Point positionMund2,
          int slideValue1, int slideValue2) {
    kopiereBilderUndPunkteZurWeiterenBearbeitung(bild1, bild2,
            positionLinkesAuge1, positionRechtesAuge1, positionMund1,
            positionLinkesAuge2, positionRechtesAuge2, positionMund2);
    macheDieBilderDeckungsgleich();
    ueberblendeDieBilder(slideValue1, slideValue2);
  }

  public void kopiereBilderUndPunkteZurWeiterenBearbeitung(BufferedImage bild1,
    BufferedImage bild2, Point positionLinkesAuge1,
    Point positionRechtesAuge1, Point positionMund1,
    Point positionLinkesAuge2, Point positionRechtesAuge2,
    Point positionMund2) {
    // Klone Bilder
    imageBuffer1 = new BufferedImage(bild1.getWidth(), bild1.getHeight(),
      bild1.getType());
    imageBuffer1.setData(bild1.getData());
    imageBuffer2 = new BufferedImage(bild2.getWidth(), bild2.getHeight(),
      bild2.getType());
    imageBuffer2.setData(bild2.getData());
    // Fülle Dreieck 1
    gesichtsDreieck1[LINKES_AUGE_X] = positionLinkesAuge1.x;
    gesichtsDreieck1[LINKES_AUGE_Y] = positionLinkesAuge1.y;
    gesichtsDreieck1[RECHTES_AUGE_X] = positionRechtesAuge1.x;
    gesichtsDreieck1[RECHTES_AUGE_Y] = positionRechtesAuge1.y;
    gesichtsDreieck1[MUND_X] = positionMund1.x;
    gesichtsDreieck1[MUND_Y] = positionMund1.y;
    // Fülle Dreieck 2
    gesichtsDreieck2[LINKES_AUGE_X] = positionLinkesAuge2.x;
    gesichtsDreieck2[LINKES_AUGE_Y] = positionLinkesAuge2.y;
    gesichtsDreieck2[RECHTES_AUGE_X] = positionRechtesAuge2.x;
    gesichtsDreieck2[RECHTES_AUGE_Y] = positionRechtesAuge2.y;
    gesichtsDreieck2[MUND_X] = positionMund2.x;
    gesichtsDreieck2[MUND_Y] = positionMund2.y;
  }

  public void macheDieBilderDeckungsgleich() {
    // Lösche allfällige alte Transformationen
    transformationBild1 = new AffineTransform();
    transformationBild2 = new AffineTransform();
    // Führe Transformationen durch
    berechneTransformationen();
    wendeTransformationenAn();
  }

  private void berechneTransformationen() {
    // Die finale Transformation setzt sich aus folgenden Teiltransformationen
    // zusammen:
    verschiebeLinkeAugenZumUrsprung();
    dreheRechteAugenAufXAchse();
    legeXAchsenRechteAugenUndYAchsenMundAufeinander();
    legeXAchsenMundAufeinander();
    zentriereGesichterInBildmitte();
  }

  private void verschiebeLinkeAugenZumUrsprung() {
    // Bild 1
    AffineTransform teilTransformation = new AffineTransform();
    teilTransformation.translate(-gesichtsDreieck1[LINKES_AUGE_X],
      -gesichtsDreieck1[LINKES_AUGE_Y]);
    transformationBild1.preConcatenate(teilTransformation);
    teilTransformation.transform(gesichtsDreieck1, 0, gesichtsDreieck1, 0, 3);
    // Bild 2
    teilTransformation = new AffineTransform();
    teilTransformation.translate(-gesichtsDreieck2[LINKES_AUGE_X],
      -gesichtsDreieck2[LINKES_AUGE_Y]);
    transformationBild2.preConcatenate(teilTransformation);
    teilTransformation.transform(gesichtsDreieck2, 0, gesichtsDreieck2, 0, 3);
  }

  private void dreheRechteAugenAufXAchse() {
    // Bild 1
    AffineTransform teilTransformation = new AffineTransform();
    teilTransformation.rotate(-Math.atan((gesichtsDreieck1[RECHTES_AUGE_Y] -
      gesichtsDreieck1[LINKES_AUGE_Y]) / (gesichtsDreieck1[RECHTES_AUGE_X] -
      gesichtsDreieck1[LINKES_AUGE_X])));
    transformationBild1.preConcatenate(teilTransformation);
    teilTransformation.transform(gesichtsDreieck1, 0, gesichtsDreieck1, 0, 3);
    // Bild 2
    teilTransformation = new AffineTransform();
    teilTransformation.rotate(-Math.atan((gesichtsDreieck2[RECHTES_AUGE_Y] -
      gesichtsDreieck2[LINKES_AUGE_Y]) / (gesichtsDreieck2[RECHTES_AUGE_X] -
      gesichtsDreieck2[LINKES_AUGE_X])));
    transformationBild2.preConcatenate(teilTransformation);
    teilTransformation.transform(gesichtsDreieck2, 0, gesichtsDreieck2, 0, 3);
  }

  private void legeXAchsenRechteAugenUndYAchsenMundAufeinander() {
    // Da die Teiltransformationen der einzelnen Bilder voneinander abhängen,
    // dürfen die Punkte erst zum Schluss für den nächsten Schritt 
    // transformiert werden
    // Bild 1
    AffineTransform teilTransformation1 = new AffineTransform();
    teilTransformation1.scale((gesichtsDreieck1[RECHTES_AUGE_X] +
      gesichtsDreieck2[RECHTES_AUGE_X]) / (2.0 *
      gesichtsDreieck1[RECHTES_AUGE_X]), (gesichtsDreieck1[MUND_Y] +
      gesichtsDreieck2[MUND_Y]) / (2.0 * gesichtsDreieck1[MUND_Y]));
    transformationBild1.preConcatenate(teilTransformation1);
    // Bild 2
    AffineTransform teilTransformation2 = new AffineTransform();
    teilTransformation2.scale((gesichtsDreieck2[RECHTES_AUGE_X] +
      gesichtsDreieck1[RECHTES_AUGE_X]) / (2.0 *
      gesichtsDreieck2[RECHTES_AUGE_X]), (gesichtsDreieck2[MUND_Y] +
      gesichtsDreieck1[MUND_Y]) / (2.0 * gesichtsDreieck2[MUND_Y]));
    transformationBild2.preConcatenate(teilTransformation2);
    // Transformiere alle Punkte
    teilTransformation1.transform(gesichtsDreieck1, 0, gesichtsDreieck1, 0, 3);
    teilTransformation2.transform(gesichtsDreieck2, 0, gesichtsDreieck2, 0, 3);
  }

  private void legeXAchsenMundAufeinander() {
    // Da die Teiltransformationen der einzelnen Bilder voneinander abhängen,
    // dürfen die Punkte erst zum Schluss für den nächsten Schritt 
    // transformiert werden
    // Bild 1
    AffineTransform teilTransformation1 = new AffineTransform();
    teilTransformation1.shear((gesichtsDreieck2[MUND_X] -
      gesichtsDreieck1[MUND_X]) / (2.0 * gesichtsDreieck1[MUND_Y]), 0.0);
    transformationBild1.preConcatenate(teilTransformation1);
    // Bild 2
    AffineTransform teilTransformation2 = new AffineTransform();
    teilTransformation2.shear((gesichtsDreieck1[MUND_X] -
      gesichtsDreieck2[MUND_X]) / (2.0 * gesichtsDreieck2[MUND_Y]), 0.0);
    transformationBild2.preConcatenate(teilTransformation2);
    // Transformiere alle Punkte
    teilTransformation1.transform(gesichtsDreieck1, 0, gesichtsDreieck1, 0, 3);
    teilTransformation2.transform(gesichtsDreieck2, 0, gesichtsDreieck2, 0, 3);
  }

  private void zentriereGesichterInBildmitte() {
    // Bild 1 & 2
    AffineTransform teilTransformation = new AffineTransform();
    teilTransformation.translate((BILD_BREITE - (gesichtsDreieck1[RECHTES_AUGE_X] -
      gesichtsDreieck1[LINKES_AUGE_X])) / 2.0, (BILD_HOEHE -
      (gesichtsDreieck1[MUND_Y] - gesichtsDreieck1[LINKES_AUGE_Y])) / 2.0);
    transformationBild1.preConcatenate(teilTransformation);
    transformationBild2.preConcatenate(teilTransformation);
  }

  private void wendeTransformationenAn() {
    // Die Transformation funktioniert nur auf Graphics2D Objekten!
    // Erzeuge dazu zuerst zwei leere ArbeitsImageBuffer der richtigen Grösse
    BufferedImage arbeitsimageBuffer1 = new BufferedImage(imageBuffer1.getWidth(),
      imageBuffer1.getHeight(), imageBuffer1.getType());
    BufferedImage arbeitsimageBuffer2 = new BufferedImage(imageBuffer2.getWidth(),
      imageBuffer2.getHeight(), imageBuffer2.getType());
    // Verknüpfe die ArbeitsImageBuffer mit Graphics2D Objekten
    Graphics2D arbeitsGrafik1 = (Graphics2D) arbeitsimageBuffer1.getGraphics();
    Graphics2D arbeitsGrafik2 = (Graphics2D) arbeitsimageBuffer2.getGraphics();
    // Initialisiere die Graphics2D Objekte mit der gewünschten Transformation
    arbeitsGrafik1.setTransform(transformationBild1);
    arbeitsGrafik2.setTransform(transformationBild2);
    // Die beiden zu transformierenden ImageBuffer werden in die Grafics2D 
    // Objekte eingelesen, die Transformationen wird durchgeführt und die 
    // Resultate werden in die verknüpften ArbeitsImageBuffer geschrieben
    arbeitsGrafik1.drawImage(imageBuffer1, 0, 0, null);
    arbeitsGrafik2.drawImage(imageBuffer2, 0, 0, null);
    // Lösche die Graphics2D Objekte wieder
    arbeitsGrafik1.dispose();
    arbeitsGrafik2.dispose();
    // Kopiere die Resultate der ArbeitsImageBuffer in die ImageBuffer
    imageBuffer1 = new BufferedImage(arbeitsimageBuffer1.getWidth(),
      arbeitsimageBuffer1.getHeight(), arbeitsimageBuffer1.getType());
    imageBuffer1.setData(arbeitsimageBuffer1.getData());
    imageBuffer2 = new BufferedImage(arbeitsimageBuffer2.getWidth(),
      arbeitsimageBuffer2.getHeight(), arbeitsimageBuffer2.getType());
    imageBuffer2.setData(arbeitsimageBuffer2.getData());
  }

  private void ueberblendeDieBilder(int slideValue1, int slideValue2) {
    // Die Bilder in "imageBuffer1" und "imageBuffer2" Bilder werden pixelweise
    // gemäss der Slide Einstellungen in "imageBuffer" der Vaterklasse überblendet.
    // Die Art der Überblendung wurde durch Thomas Jampen inspiriert. 
    Color pixelFarbeImageBuffer1, pixelFarbeImageBuffer2, pixelFarbeImageBuffer;
    int morphedRed, morphedGreen, morphedBlue;
    // Bestimmt die Mitte des Bildes
    double xBildMitte = imageBuffer.getWidth() / 2;
    double yBildMitte = imageBuffer.getHeight() / 2;
    // Die Variable slideValue1 bestimmt den äusseren Kreis der Überblendung
    // Der maximale Radius beträgt die halbe Diagonale des Bildes
    int radiusAeussererBlendkreis =
            (int) (Math.sqrt(BILD_HOEHE * BILD_HOEHE + BILD_BREITE * BILD_BREITE)
            / 2.0 * (slideValue1 / 100.0));
    // Die Variable slideValue2 bestimmt den inneren Kreis der Überblendung
    // Der maximale Radius entspricht dem äusseren Kreis der Überblendung
    int radiusInnererBlendkreis =
            (int) (radiusAeussererBlendkreis * slideValue2 / 100.0);
    // Die for-Schleifen rufen nach einander jeden einzelnen Bildpixel über die
    // Koordinaten xPosition und yPosition auf.
    for (int xPosition = 0; xPosition < BILD_BREITE; xPosition++) {
      for (int yPosition = 0; yPosition < BILD_HOEHE; yPosition++) {
        // Distanz des Bildpixels (xPosition, yPosition) zum Bildmittelpunkt
        double abstandMittelpunkt =
                Point.distance(xPosition, yPosition, xBildMitte, yBildMitte);
        // Legt fest, welches Bild wie stark berücksichtigt wird, 1 bedeutet
        // 100%, 0 entsprechend 0%
        double anteilImageBuffer1 = 1; // dummy Wert
        double anteilImageBuffer2 = 0; // dummy Wert
        // Bildpixel liegt ausserhalb äusseren Kreis der Überblendung => Bild 1
        // Bildpixel liegt innerhalb inneren Kreis der Überblendung => Bild 2
        // Dazwischen, linearer Übergang zwischen Bild 1 und Bild 2
        if (abstandMittelpunkt > radiusAeussererBlendkreis) {
          anteilImageBuffer1 = 1;
        } else if (abstandMittelpunkt < radiusInnererBlendkreis) {
          anteilImageBuffer1 = 0;
        } else {
          anteilImageBuffer1 = (abstandMittelpunkt - radiusInnererBlendkreis)
                  / (radiusAeussererBlendkreis - radiusInnererBlendkreis);
        }
        anteilImageBuffer2 = 1 - anteilImageBuffer1;
        // Pixelweise jeden RGB Farbkanal einzeln überblenden
        pixelFarbeImageBuffer1 = new Color(imageBuffer1.getRGB(xPosition,
                yPosition), true);
        pixelFarbeImageBuffer2 = new Color(imageBuffer2.getRGB(xPosition,
                yPosition), true);
        morphedRed = (int) (anteilImageBuffer1
                * pixelFarbeImageBuffer1.getRed()
                + anteilImageBuffer2
                * pixelFarbeImageBuffer2.getRed());
        // Aufgrund der Rundungsfehler können unzulässige Farbwerte über    255
        // entstehen, diese werden auf den Farbwert 255 gesetzt
        if (morphedRed > 255) {
          morphedRed = 255;
        }
        morphedGreen = (int) (anteilImageBuffer1
                * pixelFarbeImageBuffer1.getGreen()
                + anteilImageBuffer2
                * pixelFarbeImageBuffer2.getGreen());
        if (morphedGreen > 255) {
          morphedGreen = 255;
        }
        morphedBlue = (int) (anteilImageBuffer1
                * pixelFarbeImageBuffer1.getBlue()
                + anteilImageBuffer2
                * pixelFarbeImageBuffer2.getBlue());
        if (morphedBlue > 255) {
          morphedBlue = 255;
        }
        // RGB Farbkanäle wieder zusammensetzen
        pixelFarbeImageBuffer = new Color(morphedRed,
                morphedGreen, morphedBlue);
        imageBuffer.setRGB(xPosition, yPosition,
                pixelFarbeImageBuffer.getRGB());
      }
    }
    repaint();
  }

  public void loescheBild() {
    // Methode der Vaterklasse
    zeichneStartBild();
  }
}
